/*
    KrossWordPuzzle - A crossword playing game and editor for KDE4.
    This file contains code for base drawing / interaction / data of crossword cells.
    Copyright (C) 2009     Friedrich P?lz <fpuelz@gmx.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "krosswordcell.h"
#include "krossword.h"
#include "krosswordtheme.h"
#include "krosswordrenderer.h"
#include "cells/cluecell.h"

#include <QGraphicsScene>
#include <QGraphicsSceneMouseEvent>
#include <qevent.h>
#include <kdebug.h>
#include <kdeversion.h>

#if QT_VERSION >= 0x040600
    #include <QPropertyAnimation>
    #include <QParallelAnimationGroup>
#include <QGraphicsEffect>
#endif

namespace Crossword {

#if QT_VERSION >= 0x040600
KrossWordCell::KrossWordCell( KrossWord* krossWord, CellType cellType,
                              const Coord& coord )
	: QGraphicsObject(krossWord),
#else
KrossWordCell::KrossWordCell( KrossWord* krossWord, CellType cellType,
                              const Coord& coord )
        : QGraphicsItem(krossWord),
#endif
	    m_cache(0) {
    m_krossWord = krossWord;
    m_cellType = cellType;
    m_coord = coord;
    m_highlight = false;

    
    setFlag( QGraphicsItem::ItemIsFocusable );
    setFlag( QGraphicsItem::ItemIsSelectable );
// 	setFlag( QGraphicsItem::ItemUsesExtendedStyleOption );
// 	setFlag( QGraphicsItem::ItemSendsGeometryChanges );

    setPositionFromCoordinates( false );
    
#if QT_VERSION >= 0x040600
    // Set new cells completely transparent, they will fade in when they are
    // added to the crossword by KrossWord::replaceCell.
    setOpacity( 0 );

    QGraphicsDropShadowEffect *effect = new QGraphicsDropShadowEffect;
    effect->setEnabled( false );
    setGraphicsEffect( effect );
#endif

    // Connect focus event of the cell to the crossword object
//     connect( this, SIGNAL(gotFocus(KrossWordCell*)),
// 	     krossWord, SLOT(setCurrentCell(KrossWordCell*)) );
// 	     krossWord, SLOT(focusCellChanged(KrossWordCell*)) );
}

KrossWordCell::~KrossWordCell() {
  if ( scene() )
    scene()->removeItem( this );
  
  removeSynchronization();
  delete m_cache;
}

void KrossWordCell::setCoord( Coord newCoord, bool updateInCrosswordGrid ) {
  if ( updateInCrosswordGrid ) {
    // For clue cells (and maybe others): Only remove from the old position
    // if this cell is at that position (don't remove letter cells when the clue
    // cell is hidden).
    if ( krossWord()->at(coord()) == this )
      krossWord()->removeCell( coord(), false );

    krossWord()->replaceCell( newCoord, this );
  } else
    m_coord = newCoord;

  emit cellMoved( newCoord );
}

void KrossWordCell::setPositionFromCoordinates( bool animate ) {
#if QT_VERSION < 0x040600
  Q_UNUSED( animate );
#endif

  QPointF newPos( coord().first * (krossWord()->cellSize().width() / 2),
		  coord().second * (krossWord()->cellSize().height() / 2) );
  newPos += krossWord()->topLeftCellOffset();
  
#if QT_VERSION >= 0x040600
  if ( animate && krossWord()->isAnimationTypeEnabled(AnimatePosChange) ) {
    QPropertyAnimation *moveAnim = new QPropertyAnimation( this, "pos" );
    moveAnim->setDuration( krossWord()->animationDuration() * 3 );
    moveAnim->setStartValue( pos() );   
    moveAnim->setEndValue( newPos );  
    moveAnim->setEasingCurve( QEasingCurve(QEasingCurve::InOutBack) );
    moveAnim->start( QAbstractAnimation::DeleteWhenStopped );
    connect( moveAnim, SIGNAL(finished()), this, SLOT(updateTransformOriginPoint()) );
  } else {
    setPos( newPos );
    updateTransformOriginPoint();
  }
  
#else
  setPos( newPos );
#endif
}

#if QT_VERSION >= 0x040600
void KrossWordCell::updateTransformOriginPoint() {
  setTransformOriginPoint( x() + boundingRect().width() / 2,
			   y() + boundingRect().height() / 2 );
}
#endif

void KrossWordCell::clearCache() {
#if QT_VERSION >= 0x040600
  if ( m_cache ) {
    if ( krossWord()->isAnimationTypeEnabled(AnimateTransition) ) {
      GraphicsPixmapObject *transitionItem = new GraphicsPixmapObject( *m_cache, this );
      transitionItem->setTransformationMode( Qt::SmoothTransformation );
      transitionItem->setObjectName( "transitionItem" );
      transitionItem->setZValue( 1000 );
      transitionItem->setPos( pos() );
      transitionItem->setTransformOriginPoint(
	  transitionItem->boundingRect().width() / 2,
	  transitionItem->boundingRect().height() / 2 );
      QPropertyAnimation *fadeOutAnim = new QPropertyAnimation(
	  transitionItem, "opacity" );
      fadeOutAnim->setDuration( krossWord()->animationDuration() );
      fadeOutAnim->setStartValue( 1 );
      fadeOutAnim->setEndValue( 0 );
      fadeOutAnim->start( QAbstractAnimation::DeleteWhenStopped );
      // TODO: Does GraphicsPixmapObject delete m_cache (it's pixmap)?
      connect( fadeOutAnim, SIGNAL(finished()), transitionItem, SLOT(deleteLater()) );
      connect( this, SIGNAL(rotationChanged()),
	       transitionItem, SLOT(setRotationOfParent()) );
      connect( this, SIGNAL(scaleChanged()),
	       transitionItem, SLOT(setScaleOfParent()) );
    } else
      delete m_cache;
  }  
#else
  delete m_cache;
#endif

  m_cache = NULL;
}

void KrossWordCell::synchronizeWith( const KrossWordCellList &cellList,
				     SyncMethods syncMethods,
				     SyncCategory syncCategory ) {
    foreach ( KrossWordCell *cell, cellList )
	synchronizeWith( cell, syncMethods, syncCategory );
}

void KrossWordCell::synchronizeWith( KrossWordCell* cell,
				     SyncMethods syncMethods,
				     SyncCategory syncCategory ) {
    Q_ASSERT( cell );
    if ( cell == this )
	return; // Don't sync cell to itself

    bool syncContent = syncMethods.testFlag( SyncContent );
    bool syncSelection = syncMethods.testFlag( SyncSelection );
    bool contentSynced = isSynchronizedWith( cell, SyncContent );
    bool selectionSynced = isSynchronizedWith( cell, SyncSelection );
    if ( (contentSynced || !syncContent) && (selectionSynced || !syncSelection) )
	return;

//     qDebug() << "Synchronize" << coord() << "with" << cell->coord()
// 	<< "syncMethods =" << syncMethods << "syncCategory =" << syncCategory;
    if ( syncContent && !contentSynced ) {
// 	kDebug() << "Connected content sync" << coord() << "with" << cell->coord();
	connect( cell, SIGNAL(currentLetterChanged(LetterCell*,QChar)),
		 this, SLOT(setCurrentLetterSlot(LetterCell*,QChar)) );
	connect( this, SIGNAL(currentLetterChanged(LetterCell*,QChar)),
		 cell, SLOT(setCurrentLetterSlot(LetterCell*,QChar)) );
    }

    if ( syncSelection && !selectionSynced ) {
	connect( cell, SIGNAL(gotFocus(KrossWordCell*)),
		 this, SLOT(setFocusSlot(KrossWordCell*)) );
	connect( this, SIGNAL(gotFocus(KrossWordCell*)),
		 cell, SLOT(setFocusSlot(KrossWordCell*)) );
    }

    if ( m_synchronizedCells[syncCategory].contains(cell) ) {
	Q_ASSERT( cell->m_synchronizedCells[syncCategory].contains(this) );
	cell->m_synchronizedCells[syncCategory][this] |= syncMethods;
	this->m_synchronizedCells[syncCategory][cell] |= syncMethods;
    } else {
	cell->m_synchronizedCells[syncCategory].insert( this, syncMethods );
	this->m_synchronizedCells[syncCategory].insert( cell, syncMethods );
    }
}

bool KrossWordCell::isSynchronizedWith( KrossWordCell* cell,
					SyncCategory syncCategory,
					SyncMethods syncMethods ) {
    // Return true, if at least one of syncMethods is synced in syncCategory
    return m_synchronizedCells.value( syncCategory ).contains( cell )
	&& ( m_synchronizedCells.value(syncCategory).value(cell) & syncMethods) != 0;
}

bool KrossWordCell::isSynchronizedWith( KrossWordCell* cell,
					SyncMethods syncMethods,
					SyncCategories syncCategories ) {
    QList< SyncCategory > cats = allSynchronizationCategories();
    foreach ( SyncCategory cat, cats ) {
	if ( !syncCategories.testFlag(cat) )
	    continue;

	if ( isSynchronizedWith(cell, cat, syncMethods) )
	    return true;
    }

    return false;
}

bool KrossWordCell::removeSynchronizationWith( KrossWordCell* cell,
					       SyncMethods syncMethods,
					       SyncCategories syncCategories ) {
//     if ( !m_synchronizedCells.contains(cell) )
// 	return false;
//     Q_ASSERT( cell->m_synchronizedCells.contains(this) );
  if ( !isSynchronizedWith(cell) ) {
    kDebug() << "Not Synchronized";
    return false;
  }
  Q_ASSERT( cell->isSynchronizedWith(this) );

  QList< SyncCategory > cats = allSynchronizationCategories();
  bool stillSyncedContent = false, stillSyncedSelection = false;
  foreach ( SyncCategory cat, cats ) {
    if ( syncCategories.testFlag(cat) ) {
// 	    kDebug() << "Remove sync methods" << syncMethods << "from cat" << cat
// 		<< this->coord() << this->m_synchronizedCells[ cat ][cell]
// 		<< cell->coord() << cell->m_synchronizedCells[ cat ][this];

      cell->m_synchronizedCells[ cat ][this] &= ~syncMethods;
      this->m_synchronizedCells[ cat ][cell] &= ~syncMethods;

      if ( this->m_synchronizedCells[ cat ][cell] == SyncNothing ) {
	cell->m_synchronizedCells[ cat ].remove( this );
	this->m_synchronizedCells[ cat ].remove( cell );
// 		kDebug() << "  No sync methods left, removed";
      } // else
// 		kDebug() << "  Left sync methods" << this->m_synchronizedCells[ cat ][cell]
// 		    << cell->m_synchronizedCells[ cat ][cell];
    } else if ( !stillSyncedContent && m_synchronizedCells[cat].contains(cell)
		&& m_synchronizedCells[cat][cell].testFlag(SyncContent) ) {
      stillSyncedContent = true;
    } else if ( !stillSyncedSelection && m_synchronizedCells[cat].contains(cell)
		&& m_synchronizedCells[cat][cell].testFlag(SyncSelection) ) {
      stillSyncedSelection = true;
    }
  }

  if ( !stillSyncedContent && syncMethods.testFlag(SyncContent)
	    && isLetterCell() && cell->isLetterCell() ) {
//     kDebug() << "Disconnected content sync" << coord() << "with" << cell->coord();
    disconnect( cell, SIGNAL(currentLetterChanged(LetterCell*,QChar)),
		this, SLOT(setCurrentLetterSlot(LetterCell*,QChar)) );
    disconnect( this, SIGNAL(currentLetterChanged(LetterCell*,QChar)),
		cell, SLOT(setCurrentLetterSlot(LetterCell*,QChar)) );
  }

  if ( !stillSyncedSelection && syncMethods.testFlag(SyncSelection) ) {
    disconnect( cell, SIGNAL(gotFocus(KrossWordCell*)),
	      this, SLOT(setFocusSlot(KrossWordCell*)) );
    disconnect( this, SIGNAL(gotFocus(KrossWordCell*)),
	      cell, SLOT(setFocusSlot(KrossWordCell*)) );
  }

  return true;
}

QString KrossWordCell::syncInfoString() const {
  QString info = QString( "%1,%2,%3" )
    .arg( m_synchronizedCells[SolutionLetterSynchronization].count() )
    .arg( m_synchronizedCells[SameCharacterLetterSynchronization].count() )
    .arg( m_synchronizedCells[OtherSynchronization].count() );

  return info;
}

void KrossWordCell::removeSynchronization( SyncMethods syncMethods,
					   SyncCategories syncCategories ) {
  QList< SyncCategory > cats = allSynchronizationCategories();
  foreach ( SyncCategory cat, cats ) {
    if ( !syncCategories.testFlag(cat) )
      continue;

    for ( int i = m_synchronizedCells[cat].count() - 1; i >= 0; --i ) {
      KrossWordCell *cell = m_synchronizedCells[cat].keys()[ i ];
//       qDebug() << "Remove synchronization " /*<< ((SolutionLetterCell*)this)->correctLetter()*/
// 	  << "of cell at" << coord() << "with cell at" << cell->coord();
      removeSynchronizationWith( cell, syncMethods, cat );
    }
  }
}

void KrossWordCell::setFocusSlot( KrossWordCell* cell ) {
    Q_UNUSED( cell ); // TODO: remove cell parameter?
//     kDebug() << "SET FOCUS" << this->coord() << "hasFocus()?" << hasFocus();

    setFocus(); // TEST
}

void KrossWordCell::deleteAndRemoveFromSceneLater() {
  if ( scene() )
    scene()->removeItem( this );
  deleteLater();
}

QRectF KrossWordCell::boundingRect() const {
//     kDebug() << "bounding size =" << krossWord()->cellSize();
//     qreal penWidth = 1;
//     return QRectF( pos().x() - penWidth / 2, pos().y() - penWidth / 2,
// 		   krossWord()->cellSize().width() + penWidth,
// 		   krossWord()->cellSize().height() + penWidth );
    return QRectF( pos(), krossWord()->cellSize() );
}

// void KrossWordCell::drawBackground( QPainter* p, const QStyleOptionGraphicsItem* option ) {
//     Q_UNUSED( p );
//     Q_UNUSED( option );
//     // TODO Get colors from KrossWord, to make them configurable
// //     if ( option->state.testFlag( QStyle::State_HasFocus ) )
// // 	p->fillRect( option->rect, QColor(100, 100, 200) );
// //     else
// //         p->fillRect( option->rect, Qt::white );
// //
// //     p->setPen( Qt::black );
// //     p->drawRect( option->rect );
// }

QVariant KrossWordCell::itemChange( QGraphicsItem::GraphicsItemChange change,
                                    const QVariant& value ) {
    if ( change == ItemSelectedChange )
        update();
    return value;
}

// void KrossWordCell::mouseMoveEvent( QGraphicsSceneMouseEvent* event ) {
//     if ( !krossWord()->isInteractive() )
// 	return;
//
//     krossWord()->emitMouseMoved( event->scenePos(), event->button(), this );
//
//     QGraphicsItem::mouseMoveEvent( event );
// }

void KrossWordCell::mousePressEvent( QGraphicsSceneMouseEvent* event ) {
//     if ( event->button() != Qt::LeftButton )
//  event->ignore();
//     else {
//  event->accept();
//  kDebug() << "Cell clicked at" << coord() << "Type =" << cellType();
//     }
    if ( !krossWord()->isInteractive() ) {
// 	kDebug() << "KROSSWORDCELL MOUSEPRESS NOT INTERACTIVE" << this->coord();
	QGraphicsItem::mousePressEvent( event );
	return;
    } 

    krossWord()->emitMousePressed( event->scenePos(), event->button(), this );

    if ( event->button() == Qt::RightButton ) {
	if ( isLetterCell() || cellType() == EmptyCellType ) {
	    krossWord()->setHighlightedClue( NULL );
	    setHighlight();
	}
	krossWord()->emitCustomContextMenuRequested( event->scenePos(), this );
	event->accept();
    } else {
	event->ignore();
	QGraphicsItem::mousePressEvent( event );
    }
}

void KrossWordCell::mouseReleaseEvent( QGraphicsSceneMouseEvent* event ) {
    QGraphicsItem::mouseReleaseEvent( event );
}

void KrossWordCell::focusInEvent( QFocusEvent* event ) {
//     kDebug() << "KROSSWORDCELL IN AND EMITTING gotFocus()" << this->coord();
    clearCache();
    QGraphicsItem::focusInEvent( event );
//     krossWord()->setCurrentCell( this ); // is already connected to signal gotFocus() emitted below
    update();

    krossWord()->setCurrentCell( this );
    emit gotFocus( this );

#if QT_VERSION >= 0x040600
    QParallelAnimationGroup *animGroup = new QParallelAnimationGroup;
    if ( graphicsEffect() ) {
      QGraphicsDropShadowEffect *effect =
	  static_cast< QGraphicsDropShadowEffect* >( graphicsEffect() );
      if ( effect ) {
	effect->setEnabled( true );
	if ( krossWord()->isAnimationTypeEnabled(AnimateTransition) ) {
	  QPropertyAnimation *fadeColorAnim = new QPropertyAnimation( effect, "color" );
	  fadeColorAnim->setDuration( krossWord()->animationDuration() );
	  fadeColorAnim->setStartValue( effect->color() );
	  fadeColorAnim->setEndValue( krossWord()->theme()->glowFocusColor() );
// 	  fadeColorAnim->start( QAbstractAnimation::DeleteWhenStopped );
	  animGroup->addAnimation( fadeColorAnim );
	} else
	  effect->setColor( krossWord()->theme()->glowFocusColor() );
	setZValue( 5 );
      }
    }
    
    if ( krossWord()->isAnimationTypeEnabled(AnimateFocusIn) ) {
      QPropertyAnimation *scaleAnim = new QPropertyAnimation( this, "scale" );
      scaleAnim->setDuration( krossWord()->animationDuration() * 2 ); 
      scaleAnim->setStartValue( 1.0 );
      scaleAnim->setKeyValueAt( 0.5, 1.2 );
      scaleAnim->setEndValue( 1.0 );
      scaleAnim->setEasingCurve(QEasingCurve(QEasingCurve::InOutBack));
//       scaleAnim->start( QAbstractAnimation::DeleteWhenStopped );
      animGroup->addAnimation( scaleAnim );
    }

    if ( animGroup->animationCount() > 0 )
      animGroup->start( QAbstractAnimation::DeleteWhenStopped );
    else
      delete animGroup;
#endif
}

void KrossWordCell::focusOutEvent( QFocusEvent* event ) {
//     if ( scene() && scene()->hasFocus() )
	// scene still has focus
//     else
	// focus not any longer in the scene or no scene

#if QT_VERSION >= 0x040600
    QGraphicsDropShadowEffect *effect =
	static_cast< QGraphicsDropShadowEffect* >( graphicsEffect() );
    if ( effect ) {
      if ( krossWord()->isAnimationTypeEnabled(AnimateTransition) ) {
	QPropertyAnimation *fadeColorAnim = new QPropertyAnimation( effect, "color" );
	fadeColorAnim->setDuration( krossWord()->animationDuration() );
	fadeColorAnim->setStartValue( effect->color() );
	fadeColorAnim->setEndValue( krossWord()->theme()->glowColor() );
	fadeColorAnim->start( QAbstractAnimation::DeleteWhenStopped );
      } else
	effect->setColor( krossWord()->theme()->glowColor() );
      setZValue( 0 );
    }
#endif

//     m_highlight = false;
//     kDebug() << "KROSSWORDCELL OUT" << this->coord();
    clearCache();
    update();
    QGraphicsItem::focusOutEvent( event );
}

void KrossWordCell::setHighlight( bool enable ) {
#if QT_VERSION >= 0x040600
  if ( enable ) {
    QGraphicsDropShadowEffect *effect =
	static_cast< QGraphicsDropShadowEffect* >( graphicsEffect() );
    if ( effect ) {
      effect->setEnabled( true );
      effect->setOffset( 0 );
      effect->setColor( krossWord()->theme()->glowColor() );

      if ( krossWord()->isAnimationTypeEnabled(AnimateTransition) ) {
	QPropertyAnimation *blurAnim = new QPropertyAnimation( effect, "blurRadius" );
	blurAnim->setDuration( krossWord()->animationDuration() );
	blurAnim->setStartValue( 0 );
	blurAnim->setEndValue( 10 );
	blurAnim->start( QAbstractAnimation::DeleteWhenStopped );
      } else
	effect->setBlurRadius( 10 );
    }
  } else if ( graphicsEffect() )
    graphicsEffect()->setEnabled( false );
#endif
  
  m_highlight = enable;
  clearCache();
  update();
}


EmptyCell::EmptyCell( KrossWord* krossWord, Coord coord )
		      : KrossWordCell( krossWord, EmptyCellType, coord ) {
    setFlag( QGraphicsItem::ItemIsFocusable, krossWord->isEditable() );
    setFlag( QGraphicsItem::ItemIsSelectable, krossWord->isEditable() );

#if QT_VERSION >= 0x040600
    setFlag( QGraphicsItem::ItemHasNoContents, !krossWord->isEditable() );
#endif
}

LetterCell* EmptyCell::toLetterCell( const QChar& correctContent )
{
    Q_ASSERT( !krossWord()->letterContentToClueNumberMapping().isEmpty() );

    int clueNumber = krossWord()->letterContentToClueNumberMapping()
	    .indexOf( correctContent ) + 1;
    ClueCell *numberClue = new ClueCell( krossWord(), coord(),
	    Qt::Horizontal, OnClueCell,
	    clueNumber <= 0 ? QString() : QString::number(clueNumber), correctContent );
    LetterCell *letter = new LetterCell( krossWord(), coord(), numberClue );

    krossWord()->replaceCell( coord(), letter, false );
    deleteLater();

    return letter;
}

void EmptyCell::focusInEvent( QFocusEvent* event )
{
    krossWord()->setHighlightedClue( NULL );
    if ( krossWord()->isEditable() )
	setHighlight();
    KrossWordCell::focusInEvent( event );
}

void EmptyCell::focusOutEvent( QFocusEvent* event )
{
    // Only remove highlight if the scene still has focus
    if ( scene() && scene()->hasFocus() )
	setHighlight( false );
    KrossWordCell::focusOutEvent( event );
}

void EmptyCell::keyPressEvent( QKeyEvent* event )
{
    Q_UNUSED( event );
    // TODO
//     if ( krossWord()->isEditable() && event->text().length() >= 1
// 	    && krossWord()->crosswordTypeInfo().isCharacterLegal(event->text().at(0))
// 	    && krossWord()->crosswordTypeInfo().clueType == KrossWord::NumberClues1To26
// 	    && krossWord()->crosswordTypeInfo().clueMapping == KrossWord::CluesReferToCells
// 	    && krossWord()->crosswordTypeInfo().letterCellContent == KrossWord::Characters ) {
// 	QChar newChar = event->text().toUpper()[0];
// 	toLetterCell( newChar );
//     }
}

void EmptyCell::mousePressEvent( QGraphicsSceneMouseEvent* event ) {
    if ( !krossWord()->isEditable() ) {
	event->ignore();
	return;
    }

    if ( event->button() == Qt::LeftButton ) {
	setHighlight();
	event->accept();
    } else
	event->ignore();

    KrossWordCell::mousePressEvent( event );
}

void EmptyCell::drawBackground( QPainter* p, const QStyleOptionGraphicsItem* option ) {
    if ( krossWord()->isEditable() ) {
	if ( isHighlighted() ) {
	    p->fillRect( option->rect,krossWord()->theme()->selectionColor() );
	    p->drawRect( option->rect.adjusted(0, 0, -1, -1) );
	} else {
	    p->fillRect( option->rect, krossWord()->theme()->emptyCellColor() );
	    p->drawRect( option->rect.adjusted(0, 0, -1, -1) );
	}
    }
}

void EmptyCell::drawBackgroundForPrinting( QPainter* p, const QStyleOptionGraphicsItem* option ) {
    p->fillRect( option->rect, krossWord()->emptyCellColorForPrinting() );
}

void KrossWordCell::paint( QPainter* painter, const QStyleOptionGraphicsItem* option, QWidget* widget ) {
    Q_UNUSED( widget );

//     drawBackground( painter, option );
//     drawForeground( painter, option );

    if ( krossWord()->isDrawingForPrinting() ) {
	drawBackgroundForPrinting( painter, option );
	drawForegroundForPrinting( painter, option );
    } else {
	if ( !m_cache ) {
#if QT_VERSION >= 0x040600
	    // TODO
// 	    kDebug() << option->exposedRect;
	    qreal levelOfDetail = QStyleOptionGraphicsItem::levelOfDetailFromTransform( sceneTransform() );
#else
	    qreal levelOfDetail = option->levelOfDetail;
#endif
// kDebug() << levelOfDetail;
	    m_cache = new QPixmap( option->rect.size() * levelOfDetail );
	    m_cache->fill( Qt::transparent );
	    QPainter p( m_cache );
	    p.translate( -option->rect.topLeft() );

	    QStyleOptionGraphicsItem *scaledOption = new QStyleOptionGraphicsItem( *option );
	    scaledOption->rect.setWidth( option->rect.width() * levelOfDetail );
	    scaledOption->rect.setHeight( option->rect.height() * levelOfDetail );
	    drawBackground( &p, scaledOption );
	    drawForeground( &p, scaledOption );
	    p.end();
	}

	painter->setRenderHints( QPainter::HighQualityAntialiasing | QPainter::Antialiasing
		    | QPainter::SmoothPixmapTransform | QPainter::TextAntialiasing );
	painter->drawPixmap( option->rect, *m_cache );
    }
}

bool KrossWordCell::isHighlighted() const {
  return m_highlight && !krossWord()->isDrawingForPrinting();
}


// Sorting functions:
bool lessThanCellType( const KrossWordCell* cell1, const KrossWordCell* cell2 ) {
    return cell1->cellType() < cell2->cellType();
}

bool greaterThanCellType( const KrossWordCell* cell1, const KrossWordCell* cell2 ) {
    return cell1->cellType() > cell2->cellType();
}

}; // namespace Crossword
