/*
    KrossWordPuzzle - A crossword playing game and editor for KDE4.
    This file contains code for base drawing / interaction / data of crossword cells.
    Copyright (C) 2009     Friedrich P?lz <fpuelz@gmx.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef KROSSWORDCELL_H
#define KROSSWORDCELL_H


#include <QString>
#include <QPair>
#include <QPainter>
#include <qstyleoption.h>

#include <KDebug>
#include <kdeversion.h>

#include "global.h"

#include <qnamespace.h>

#if QT_VERSION >= 0x040600
#include <QGraphicsObject>
#else
#include <QGraphicsItem>
#endif

namespace Crossword {

class KrossWord;
class KrossWordCell;
class EmptyCell;
class ClueCell;
class DoubleClueCell;
class LetterCell;
class SolutionLetterCell;
class ImageCell;

#if QT_VERSION >= 0x040600
class GraphicsPixmapObject : public QObject, public QGraphicsPixmapItem {
  Q_OBJECT
  Q_PROPERTY(qreal opacity READ opacity WRITE setOpacity NOTIFY opacityChanged)

  public:
    GraphicsPixmapObject( const QPixmap &pixmap, QGraphicsItem *parent = 0 )
	: QObject(0), QGraphicsPixmapItem( pixmap, parent ) {};

  public slots:
    void setRotationOfParent() {
	setRotation( parentItem()->rotation() ); };
    void setScaleOfParent() {
	setScale( parentItem()->scale() ); };
};
#endif

#if QT_VERSION >= 0x040600
class KrossWordCell : public QGraphicsObject {
#else
class KrossWordCell : public QObject, public QGraphicsItem {
#endif
    friend class KrossWord;
    friend class DoubleClueCell; // To be able to set m_coord of the child clue cells in the constructor
    Q_OBJECT
#if QT_VERSION >= 0x040600
    Q_INTERFACES( QGraphicsItem )
#endif

    public:
	KrossWordCell( KrossWord *krossWord, CellType cellType,
		       const Coord &coord );
	virtual ~KrossWordCell();

	/** For qgraphicsitem_cast. */
	enum { Type = UserType + 1 };
	virtual int type() const { return Type; };

	CellType cellType() const { return m_cellType; };
	bool isType( CellType cellType ) const {
	    return m_cellType == cellType; };
	virtual bool isLetterCell() const { return false; };
	Coord coord() const { return m_coord; };

	/** Returns the KrossWord object, to which this cell belongs. */
	KrossWord *krossWord() const { return m_krossWord; };

	/** Returns a list with all synchronization categories. */
	QList< SyncCategory > allSynchronizationCategories() const {
	    return QList< SyncCategory >() << OtherSynchronization
			    << SolutionLetterSynchronization
			    << SameCharacterLetterSynchronization;
	};

	const QHash< SyncCategory, QHash< KrossWordCell*,
						     SyncMethods > >&
	    synchronizationByCategory() const { return m_synchronizedCells; };
	QString syncInfoString() const;

	/** Checks if this cell is synced by one of @p syncMethods with
	* @p cell in one of @p syncCategories.
	*  @returns True, if at least one of @p syncMethods is synced in
	* one of @p syncCategories.
	* @returns False, if none of @p syncMethods are synced in any of
	* @p syncCategories. */
	bool isSynchronizedWith( KrossWordCell *cell,
				 SyncMethods syncMethods =
				    SyncAll,
				 SyncCategories syncCategories =
				    AllSyncCategories );

	/** Checks if this cell is synced by one of @p syncMethods with
	* @p cell in @p syncCategory.
	*  @returns True, if at least one of @p syncMethods is synced in
	* category @p syncCategory.
	* @returns False, if none of @p syncMethods are synced in category
	* @p syncCategory. */
	bool isSynchronizedWith( KrossWordCell *cell, SyncCategory syncCategory,
				 SyncMethods syncMethods = SyncAll );

	void synchronizeWith( KrossWordCell *cell,
			      SyncMethods syncMethods = SyncAll,
			      SyncCategory syncCategory = OtherSynchronization );
	void synchronizeWith( const KrossWordCellList &cellList,
			      SyncMethods syncMethods = SyncAll,
			      SyncCategory syncCategory = OtherSynchronization );
	bool removeSynchronizationWith( KrossWordCell *cell,
					SyncMethods syncMethods = SyncAll,
					SyncCategories syncCategories = AllSyncCategories );
	void removeSynchronization( SyncMethods syncMethods = SyncAll,
				    SyncCategories syncCategories = AllSyncCategories );

	virtual void setHighlight( bool enable = true );
	bool isHighlighted() const;

	void clearCache();

// 	static QString displayStringFromCellType( CellType cellType );
// 	static QString stringFromCellType( CellType cellType );
// 	static CellType cellTypeFromString( const QString &sCellType );

    signals:
	void gotFocus( KrossWordCell *cell );
	void cellMoved( const Coord &newCoord );

    public slots:
	void setFocusSlot( KrossWordCell *cell );
	void deleteAndRemoveFromSceneLater();

    protected slots:
#if QT_VERSION >= 0x040600
	void updateTransformOriginPoint();
#endif

    protected:
	virtual void setPositionFromCoordinates( bool animate = true );
	void setCoord( Coord coord, bool updateInCrosswordGrid = true );

	// Overloaded methods
        virtual QRectF boundingRect() const;
// 	virtual void mouseMoveEvent( QGraphicsSceneMouseEvent* event );
        virtual void mousePressEvent( QGraphicsSceneMouseEvent* event );
        virtual void mouseReleaseEvent( QGraphicsSceneMouseEvent* event );
        virtual void focusInEvent( QFocusEvent* event );
	virtual void focusOutEvent( QFocusEvent* event );
	virtual QVariant itemChange( GraphicsItemChange change, const QVariant& value );
        virtual void paint( QPainter* painter,
			    const QStyleOptionGraphicsItem* option,
			    QWidget* widget = 0 );

	// Virtual methods
	virtual void drawBackground( QPainter*, const QStyleOptionGraphicsItem* ) { };
	virtual void drawForeground( QPainter*, const QStyleOptionGraphicsItem* ) { };
	virtual void drawBackgroundForPrinting( QPainter*, const QStyleOptionGraphicsItem* ) { };
	virtual void drawForegroundForPrinting( QPainter*, const QStyleOptionGraphicsItem* ) { };

// 	void updatePixmap() {
// 	    QPixmap pixmap( boundingRect().size().toSize() );
// 	    QPainter p( &pixmap );
// 	    draw( &p, boundingRect().toAlignedRect() );
// 	    p.end();
//
// // 	    setPixmap( pixmap );
// 	};

	KrossWord *m_krossWord;

    private:
	Coord m_coord;
	QHash< SyncCategory, QHash< KrossWordCell*, SyncMethods > > m_synchronizedCells;
	CellType m_cellType;
	bool m_highlight;
	QPixmap *m_cache;
};


class EmptyCell : public KrossWordCell {
    friend class KrossWord;
    Q_OBJECT

    public:
	EmptyCell( KrossWord *krossWord, Coord coord );

	/** For qgraphicsitem_cast. */
	enum { Type = UserType + 2 };
	virtual int type() const { return Type; };

	LetterCell *toLetterCell( const QChar &correctContent = ' ' );

    protected:
	virtual void focusInEvent( QFocusEvent* event );
	virtual void focusOutEvent( QFocusEvent* event );
	virtual void mousePressEvent( QGraphicsSceneMouseEvent* event );
	virtual void keyPressEvent( QKeyEvent* event );

	virtual void drawBackground( QPainter *p, const QStyleOptionGraphicsItem* option );
	virtual void drawBackgroundForPrinting( QPainter *p, const QStyleOptionGraphicsItem *option );
};



inline QDebug &operator <<(QDebug debug, CellType cellType) {
  return debug << stringFromCellType( cellType );
};

inline QDebug &operator <<(QDebug debug, KrossWordCell *cell) {
  if ( !cell )
    return debug << "NULL ";

  debug << cell->cellType() << "at" << cell->coord();
  return debug.space();
};

// Sorting functions
bool lessThanCellType( const KrossWordCell *cell1, const KrossWordCell *cell2 );
bool greaterThanCellType( const KrossWordCell *cell1, const KrossWordCell *cell2 );


// Serialization
// QDataStream &operator<< ( QDataStream &s, const KrossWordCell *cell );
// QDataStream &operator<< ( QDataStream &s, const LetterCell *cell );
// QDataStream &operator<< ( QDataStream &s, const ClueCell *cell );
//
// QDataStream &operator>> ( QDataStream &s, ClueCell cell );

}; // namespace Crossword

#endif // KROSSWORDCELL_H
