/*
    KrossWordPuzzle - A crossword playing game and editor for KDE4.
    This file contains code for the main window.
    Copyright (C) 2009     Friedrich P?lz <fpuelz@gmx.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "krosswordpuzzle.h"
#include "krosswordpuzzleview.h"
#include "krosswordrenderer.h"
#include "krossworddocument.h"
#include "settings.h"
#include "crosswordxmlguiwindow.h"
#include "libraryxmlguiwindow.h"
#include "io/krosswordxmlreader.h"

// Qt UI includes
#include <QPrintDialog>
#include <QTreeWidget>

// Other Qt includes
#include <QCloseEvent>
#include <QtGui/QPrinter>

// KDE UI includes
#include <KXMLGUIFactory>
#include <KMessageBox>
#include <KConfigDialog>
#include <KPrintPreview>
#include <kdeprintdialog.h>
#include <KFileDialog>
#include <KPushButton>
#include <KStatusBar>
#include <KMenu>
#include <KMenuBar>
#include <KToolBar>
#include <KGameThemeSelector>

// KDE action includes
#include <KAction>
#include <KActionMenu>
#include <KSelectAction>
#include <KToggleAction>
#include <KActionCollection>
#include <KStandardAction>
#include <KStandardGameAction>
#include <KRecentFilesAction>

// Other KDE includes
#include <KDE/KLocale>
#include <KStandardDirs>
#include <KGameDifficulty>
#include <KRandom>
#include <KTemporaryFile>
#include <kdeversion.h>
#include <kapplication.h>
#include <knewstuff2/engine.h>
#include <KColorScheme>
#include <KShortcutsDialog>
#include <kfileplacesmodel.h>


KrossWordPuzzle::KrossWordPuzzle()
	  : KXmlGuiWindow(),
	  m_startNewPage(0),
	  m_cornerMenu(0),
	  m_mainLibrary(0),
	  m_mainCrossword(0),
	  m_loadProgressDialog(0) {
    if ( Settings::libraryDownloadSubDir().isEmpty() ) {
	Settings::setLibraryDownloadSubDir( i18n("Downloads") );
	Settings::self()->writeConfig();
    }

    setupPlaces();
    setAcceptDrops( true );
    setObjectName( "mainKrossWordPuzzle" );
    setupActions();

    setupGUI( Save | Create );

    m_startNewPage = createStartNewPage();
    setCentralWidget( m_startNewPage );
    menuBar()->hide();
    m_hiddenToolBars = toolBars();
    foreach ( KToolBar *toolBar, m_hiddenToolBars )
      toolBar->hide();
    m_mainLibrary->statusBar()->showMessage( i18n("Welcome to KrossWordPuzzle!") );

//     KrosswordRenderer::self()->setTheme( Settings::theme() ); // Load the current theme

    QString lastUnsavedFileBeforeCrash = Settings::lastUnsavedFileBeforeCrash();
    if ( !lastUnsavedFileBeforeCrash.isEmpty() ) {
// 	KSysGuard::Processes *processes = KSysGuard::Processes::getInstance();
// 	QList<KSysGuard::Process *> processlist = processes->getAllProcesses();
// 	foreach(KSysGuard::Process * process, processlist) {
// 	    QString name = process->name;
// 	}
// 	QString sPid = lastUnsavedFileBeforeCrash.mid( lastUnsavedFileBeforeCrash.lastIndexOf("PID") + 3 );
// 	sPid = sPid.left( sPid.length() - 4 ); // Remove ".tmp" from the end
// 	bool ok;
// 	long int pid = sPid.toLong( &ok );
// 	if ( ok ) {
// 	    KSysGuard::Processes *processes = KSysGuard::Processes::getInstance();
// 	    KSysGuard::Process *process = processes->getProcess( pid );
// 	    if ( process ) {
// 		kDebug() << "Game didn't crash, it's still running!";
// 	    }
// 	}

	KGuiItem restoreButton( i18n("&Restore"), KIcon("document-open"),
		i18n("Restore the automatically saved version of an edited crossword before the crash") );
	int result = KMessageBox::questionYesNo( this,
		i18n("An unsaved crossword has been found. Most likely the game crashed, sorry!\n "
		"Do you want to restore the crossword?"), QString(), restoreButton,
		KStandardGuiItem::discard() );
	if ( result == KMessageBox::Yes ) {
		loadFile( lastUnsavedFileBeforeCrash,
			  Crossword::KrossWord::KrossWordPuzzleCompressedXmlFile, true );
	} else {
		m_mainCrossword->removeTempFile( lastUnsavedFileBeforeCrash );
	}
    }
}

void KrossWordPuzzle::setupPlaces() {
  KUrl libraryUrl = KUrl( KGlobal::dirs()->saveLocation("appdata", "library") );
  KUrl templatesUrl = KUrl( KGlobal::dirs()->saveLocation("appdata", "templates") );
  
  KFilePlacesModel *placesModel = new KFilePlacesModel();
  if ( placesModel->url(placesModel->closestItem(libraryUrl)) != libraryUrl ) {
    placesModel->addPlace( i18n("Library"), KUrl(libraryUrl),
			   "favorites", KApplication::applicationName() );
    qDebug() << "ADDED LIBRARY PLACE";
  } else
    qDebug() << "LIBRARY PLACE ALREADY THERE";
  
  if ( placesModel->url(placesModel->closestItem(templatesUrl)) != templatesUrl ) {
    placesModel->addPlace( i18n("Templates"), KUrl(templatesUrl),
			   "krosswordpuzzle", KApplication::applicationName() );
    qDebug() << "ADDED TEMPLATE PLACE";
  } else
    qDebug() << "TEMPLATE PLACE ALREADY THERE" << KApplication::applicationName();
  
  delete placesModel;
}

void KrossWordPuzzle::showStatusbarGlobal( bool show ) {
// 	if ( ui_start_new.ktabwidget->currentWidget() == ui_start_new.tabCrossword )
		m_mainCrossword->statusBar()->setVisible( show );
// 	else
		m_mainLibrary->statusBar()->setVisible( show );
}

int KrossWordPuzzle::configureShortcutsGlobal() {
	KShortcutsDialog dlg( KShortcutsEditor::AllActions, KShortcutsEditor::LetterShortcutsAllowed, this );
	if ( ui_start_new.ktabwidget->currentWidget() == ui_start_new.tabCrossword )
		dlg.addCollection( m_mainCrossword->actionCollection() );
	else
		dlg.addCollection( m_mainLibrary->actionCollection() );
	return dlg.configure( true );
}

void KrossWordPuzzle::configureToolbarsGlobal() {
	if ( ui_start_new.ktabwidget->currentWidget() == ui_start_new.tabCrossword )
		m_mainCrossword->configureToolbars();
	else
		m_mainLibrary->configureToolbars();
}

const char *KrossWordPuzzle::actionName( KrossWordPuzzle::Action actionEnum ) const {
    switch ( actionEnum ) {
	case Game_Download:
		return "game_download";
	case Game_Upload:
		return "game_upload";

// 	case Options_Themes:
// 	    return "options_themes";

	case RecentTab_RecentFilesRemove:
		return "recent_files_remove";

	default:
		kWarning() << "Action enumerable not handled in switch" << actionEnum;
		return "";
    }
}

bool KrossWordPuzzle::createNewCrossWord(
		    const Crossword::CrosswordTypeInfo &crosswordTypeInfo,
		    const QSize &crosswordSize,
		    const QString& title, const QString& authors,
		    const QString& copyright, const QString& notes ) {
  if ( m_mainCrossword->createNewCrossWord(crosswordTypeInfo, crosswordSize,
		  title, authors, copyright, notes) ) {
    // HACK "Undo" setFixedHeight(0), what was called to have the crossword window
    // being created "visible" (in the library tab, but with height = 0, so it's not really visible).
    // Otherwise the graphics items somehow can't get focus...
    m_mainCrossword->setMaximumHeight( QWIDGETSIZE_MAX );
    ui_start_new.widgetCrossword->layout()->addWidget( m_mainCrossword );

    int indexCrossword = ui_start_new.ktabwidget->indexOf( ui_start_new.tabCrossword );
    ui_start_new.ktabwidget->setTabEnabled( indexCrossword, true );
    ui_start_new.ktabwidget->setCurrentIndex( indexCrossword );
    return true;
  } else
    return false;
}

bool KrossWordPuzzle::createNewCrossWordFromTemplate(
		    const QString& templateFilePath, const QString& title,
		    const QString& authors, const QString& copyright,
		    const QString& notes ) {
  if ( m_mainCrossword->createNewCrossWordFromTemplate(
		  templateFilePath, title, authors, copyright, notes) ) {
    // HACK "Undo" setFixedHeight(0), what was called to have the crossword window
    // being created "visible" (in the library tab, but with height = 0, so it's not really visible).
    // Otherwise the graphics items somehow can't get focus...
    m_mainCrossword->setMaximumHeight( QWIDGETSIZE_MAX );
    ui_start_new.widgetCrossword->layout()->addWidget( m_mainCrossword );

    int indexCrossword = ui_start_new.ktabwidget->indexOf( ui_start_new.tabCrossword );
    ui_start_new.ktabwidget->setTabEnabled( indexCrossword, true );
    ui_start_new.ktabwidget->setCurrentIndex( indexCrossword );
    return true;
  } else
    return false;
}

QWidget *KrossWordPuzzle::createStartNewPage() {
    QWidget *startNewPage = new QWidget( this );
    ui_start_new.setupUi( startNewPage );

    // Set tab  icons
    QString puzIconName = KMimeType::findByPath( "crossword.kwp", 0, true )->iconName();
    KIcon puzIcon;
    puzIcon.addPixmap( KIconLoader::global()->loadMimeTypeIcon(puzIconName, KIconLoader::Dialog) );
    int indexLibrary = ui_start_new.ktabwidget->indexOf( ui_start_new.tabLibrary );
    int indexCrossword = ui_start_new.ktabwidget->indexOf( ui_start_new.tabCrossword );
    ui_start_new.ktabwidget->setTabIcon( indexLibrary, KIcon("favorites") );
    ui_start_new.ktabwidget->setTabIcon( indexCrossword, puzIcon );
    ui_start_new.ktabwidget->setTabEnabled( indexCrossword, false );

    m_mainLibrary = new LibraryXmlGuiWindow( ui_start_new.libraryTree, this );
    m_mainLibrary->setParent( ui_start_new.widgetLibrary );
    ui_start_new.widgetLibrary->layout()->addWidget( m_mainLibrary );
//     show(); // When calling show() here, no menu is visible
    
    // Setup crossword tab
    m_mainCrossword = new CrossWordXmlGuiWindow( 0 );
    m_mainCrossword->setParent( ui_start_new.widgetCrossword );

    // HACK This is needed to have the crossword window being created "visible"
    // (in the library tab, but with height = 0, so it's not really visible).
    // Otherwise the graphics items somehow can't get focus...
    ui_start_new.widgetLibrary->layout()->addWidget( m_mainCrossword );
    m_mainCrossword->setFixedHeight( 0 );

    m_mainCrossword->krossWord()->setAnimationTypes( animationTypesFromSettings() );
    connect( m_mainCrossword, SIGNAL(loadingFileComplete(QString)),
	     this, SLOT(crosswordLoadingComplete(QString)) );
    connect( m_mainCrossword, SIGNAL(errorLoadingFile(QString)),
	     this, SLOT(crosswordErrorLoading(QString)) );
    connect( m_mainCrossword, SIGNAL(currentFileChanged(QString,QString)),
	     this, SLOT(crosswordCurrentChanged(QString,QString)) );
    connect( m_mainCrossword, SIGNAL(fileClosed(QString)),
	     this, SLOT(crosswordClosed(QString)) );
    connect( m_mainCrossword,
	     SIGNAL(modificationTypesChanged(CrossWordXmlGuiWindow::ModificationTypes)),
	     this, SLOT(crosswordModificationsChanged(CrossWordXmlGuiWindow::ModificationTypes)) );
    connect( m_mainCrossword, SIGNAL(tempAutoSaveFileChanged(QString)),
	     this, SLOT(crosswordAutoSaveFileChanged(QString)) );

    // Setup recent tab
//     ui_start_new.openRecent->setIcon( KIcon("document-open-recent") );
    // TODO: revival of the recent files action?
//     foreach ( KUrl url, m_recentFilesAction->urls() ) {
// 	QString iconName = KMimeType::findByUrl( url, 0, /*url.isLocalFile()*/true )->iconName();
// 	KIcon icon;
// 	icon.addPixmap( KIconLoader::global()->loadMimeTypeIcon(iconName, KIconLoader::Dialog) );
// 	QListWidgetItem *item = new QListWidgetItem( icon,
// 	    QString("%1\n  (%2)").arg(url.fileName()).arg(url.pathOrUrl()) );
// // 	item->setData( Qt::UserRole, url ); // QListWidgetItem doesn't store Qt::UserRole-data...
//
// 	ui_start_new.recentFiles->addItem( item );
//     }
/*
    if ( ui_start_new.recentFiles->count() == 0 ) {
	ui_start_new.openRecent->hide();
	ui_start_new.recentFiles->hide();
	ui_start_new.lblNoRecentFiles->show();

// 	ui_start_new.ktabwidget->setCurrentIndex( indexLoad );
    } else {
	ui_start_new.lblNoRecentFiles->hide();
	ui_start_new.tabRecent->layout()->removeItem( ui_start_new.spacerNoRecentFilesTop );
	ui_start_new.tabRecent->layout()->removeItem( ui_start_new.spacerNoRecentFilesBetween );
	ui_start_new.tabRecent->layout()->removeItem( ui_start_new.spacerNoRecentFilesBottom );
	ui_start_new.recentFiles->setCurrentRow( 0 );
	QListWidgetItem *mostRecentItem = ui_start_new.recentFiles->item( 0 );
	QFont font = mostRecentItem->font();
	font.setBold( true );
	mostRecentItem->setFont( font );

// 	ui_start_new.ktabwidget->setCurrentIndex( indexRecent );
    }*/

//     connect( ui_start_new.recentFiles, SIGNAL(executed(QListWidgetItem*)),
// 	    this, SLOT(recentFileExecuted(QListWidgetItem*)) );
//     connect( ui_start_new.recentFiles, SIGNAL(customContextMenuRequested(QPoint)),
// 	    this, SLOT(recentFileListContextMenuRequested(QPoint)) );
//     connect( ui_start_new.openRecent, SIGNAL(clicked()), this, SLOT(loadRecentItem()) );

	// Set a menu bar as corner widget of the main tab widget
	QWidget *w = new QWidget;
	m_cornerMenu = new KMenuBar();

	foreach( QAction *action, m_mainCrossword->menuBar()->actions() ) {
		if ( action->menu() && (action->menu()->objectName() == "game"
				|| action->menu()->objectName() == "edit"
				|| action->menu()->objectName() == "move"
				|| action->menu()->objectName() == "view") ) {
		    kDebug() << "Added menu" << action->menu()->objectName();
		    m_cornerMenu->addMenu( action->menu() );
		}
	}
	// Add the settings menu object (is there a better way?)
	foreach( QAction *action, menuBar()->actions() ) {
		if ( action->menu() && action->menu()->objectName() == "settings" ) {
			m_cornerMenu->addMenu( action->menu() );
			break;
		}
	}
	m_cornerMenu->addMenu( helpMenu() );

	QHBoxLayout *layout = new QHBoxLayout;
	layout->addWidget( m_cornerMenu, 0, Qt::AlignTop );
	w->setLayout( layout );
	ui_start_new.ktabwidget->setCornerWidget( w );

	ui_start_new.ktabwidget->setCurrentIndex( indexLibrary );
	connect( ui_start_new.ktabwidget, SIGNAL(currentChanged(int)),
			 this, SLOT(currentTabChanged(int)) );
	currentTabChanged( indexLibrary );

    return startNewPage;
}

void KrossWordPuzzle::currentTabChanged( int index ) {
    bool crosswordTabShown = index == ui_start_new.ktabwidget->indexOf(ui_start_new.tabCrossword);

    foreach ( QAction *action, m_cornerMenu->actions() ) {
	if ( action->menu() && (action->menu()->objectName() == "game"
		    || action->menu()->objectName() == "edit"
		    || action->menu()->objectName() == "move"
		    || action->menu()->objectName() == "view") ) {
	    action->setVisible( crosswordTabShown );
	}
    }

    unplugActionList( "options_list" );

    QList<QAction*> optionsList;
    QAction *seperator = new QAction( this );
    seperator->setSeparator( true );
    if ( crosswordTabShown ) {
	setCaption( m_caption, m_mainCrossword->isModified() );

	optionsList << m_mainCrossword->action(
	    m_mainCrossword->actionName(CrossWordXmlGuiWindow::Options_Dictionaries) );
	optionsList << seperator;
// 		optionsList << m_mainCrossword->action( "options_show_menubar" );
	optionsList << m_mainCrossword->toolBarMenuAction();
	optionsList << m_mainCrossword->action(
	    m_mainCrossword->actionName(CrossWordXmlGuiWindow::ShowClueDock) );
	optionsList << m_mainCrossword->action(
	    m_mainCrossword->actionName(CrossWordXmlGuiWindow::ShowUndoViewDock) );
	optionsList << m_mainCrossword->action(
	    m_mainCrossword->actionName(CrossWordXmlGuiWindow::ShowCurrentCellDock) );

	kDebug() << optionsList;
    } else { // tabLibrary
	setCaption( i18n("Library") );

	optionsList << m_mainCrossword->action(
	    m_mainCrossword->actionName(CrossWordXmlGuiWindow::Options_Dictionaries) );
	optionsList << seperator;
	optionsList << m_mainLibrary->toolBarMenuAction();
    }

    plugActionList( "options_list", optionsList );
//     kDebug() << "Options actions plugged";
}

void KrossWordPuzzle::crosswordLoadingComplete( const QString& fileName ) {
    Q_UNUSED( fileName );

    if ( !m_loadProgressDialog )
      return; // When loading a template there is no load progress dialog
    
    m_loadProgressDialog->close();

    // HACK "Redo" setFixedHeight(0), what was called to have the crossword window
    // being created "visible" (in the library tab, but with height = 0, so it's not really visible).
    // Otherwise the graphics items somehow can't get focus...
    m_mainCrossword->setMaximumHeight( QWIDGETSIZE_MAX );
    ui_start_new.widgetCrossword->layout()->addWidget( m_mainCrossword );

    int indexCrossword = ui_start_new.ktabwidget->indexOf( ui_start_new.tabCrossword );
    ui_start_new.ktabwidget->setTabEnabled( indexCrossword, true );
    ui_start_new.ktabwidget->setCurrentIndex( indexCrossword );

    m_mainLibrary->statusBar()->showMessage( i18n("Loaded '%1'", fileName) );
}


void KrossWordPuzzle::crosswordErrorLoading( const QString& fileName ) {
    m_loadProgressDialog->close();
    m_mainLibrary->statusBar()->showMessage( "Error loading file " + fileName ); // TODO get error message from signal...
}

void KrossWordPuzzle::crosswordClosed( const QString& fileName ) {
    Q_UNUSED( fileName );

    int indexCrossword = ui_start_new.ktabwidget->indexOf( ui_start_new.tabCrossword );
    int indexLibrary = ui_start_new.ktabwidget->indexOf( ui_start_new.tabLibrary );
    ui_start_new.ktabwidget->setTabEnabled( indexCrossword, false );
    ui_start_new.ktabwidget->setCurrentIndex( indexLibrary );
}

void KrossWordPuzzle::crosswordCurrentChanged( const QString& fileName,
					       const QString& oldFileName ) {
//     Q_UNUSED( fileName );
  Q_UNUSED( oldFileName );

  if ( fileName.isEmpty() ) {
    m_caption = QString();
  } else {
    stateChanged( "no_file_opened", StateReverse );

    if ( m_mainCrossword->krossWord()->title().isEmpty() )
      m_caption = displayFileName( fileName );
    else
      m_caption = m_mainCrossword->krossWord()->title();
  }

  setCaption( m_caption, m_mainCrossword->isModified() );
}

void KrossWordPuzzle::crosswordModificationsChanged(
	    CrossWordXmlGuiWindow::ModificationTypes modificationTypes ) {
//   Q_UNUSED( modificationTypes );
  int iCrossword = ui_start_new.ktabwidget->indexOf( ui_start_new.tabCrossword );
  if ( modificationTypes == CrossWordXmlGuiWindow::NoModification ) {
    ui_start_new.ktabwidget->setTabText( iCrossword, i18nc("The title for the "
	  "crossword tab with an unmodified or no crossword opened", "&Crossword") );
//     if ( ui_start_new.ktabwidget->isTabEnabled(iCrossword) )
//       ui_start_new.ktabwidget->setTabTextColor( iCrossword,
// 	  KColorScheme(QPalette::Active).foreground().color() );
//     else
//       ui_start_new.ktabwidget->setTabTextColor( iCrossword,
// 	  KColorScheme(QPalette::Disabled).foreground().color() );
  } else {
//     ui_start_new.ktabwidget->setTabTextColor( iCrossword,
// 	  KColorScheme(QPalette::Active).foreground(
// 	  KColorScheme::NeutralText).color() );
	      
    if ( modificationTypes.testFlag(CrossWordXmlGuiWindow::ModifiedCrossword) ) {
      ui_start_new.ktabwidget->setTabText( iCrossword, i18nc("The title for the "
	    "crossword tab with an edited crossword opened", "&Crossword *") );
	    
//       if ( modificationTypes.testFlag(CrossWordXmlGuiWindow::ModifiedState) )
// 	// Both edited and state changed
// 	ui_start_new.ktabwidget->setTabTextColor( iCrossword,
// 	      KColorScheme(QPalette::Active).foreground(
// 	      KColorScheme::NeutralText).color() );
//       else
// // 	// Only edited
// 	ui_start_new.ktabwidget->setTabTextColor( iCrossword,
// 	      KColorScheme(QPalette::Active).foreground(
// 	      KColorScheme::NeutralText).color() );
    } else if ( modificationTypes.testFlag(CrossWordXmlGuiWindow::ModifiedState) ) {
      ui_start_new.ktabwidget->setTabText( iCrossword, i18nc("The title for the "
	    "crossword tab with an unmodified or no crossword opened", "&Crossword") );
	    
//       // Only state changed
//       ui_start_new.ktabwidget->setTabTextColor( iCrossword,
// 	    KColorScheme(QPalette::Active).foreground(
// 	    KColorScheme::NeutralText).color() );
    }
  }

  crosswordCurrentChanged( m_mainCrossword->currentFileName(),
			   m_mainCrossword->currentFileName() );
}

void KrossWordPuzzle::crosswordAutoSaveFileChanged( const QString &fileName ) {
    Settings::setLastUnsavedFileBeforeCrash( fileName );
    Settings::self()->writeConfig();
}

QString KrossWordPuzzle::displayFileName( const QString &fileName ) {
    QString libraryDir = KGlobal::dirs()->saveLocation( "appdata", "library" );
    if ( fileName.startsWith(libraryDir) ) {
	// Cut the library path
	QString libraryFileName = fileName.mid( libraryDir.length() );
	libraryFileName.prepend( i18nc("This string is used to replace the library path "
	    "for crossword files that are in the library with a shorter user visible string, "
	    "e.g. replacing ~/.kde4/apps/krosswordpuzzle/library/crossword.kwpz with "
	    "Library/crossword.kwpz", "Library") + QDir::separator() );
	return libraryFileName;
    } else
	return fileName;
}

bool KrossWordPuzzle::isFileInLibrary( const QString& fileName ) {
    QString libraryDir = KGlobal::dirs()->saveLocation( "appdata", "library" );
    return fileName.startsWith( libraryDir );
}

void KrossWordPuzzle::loadFile( const QString& fileName ) {

    loadFile( KUrl(fileName) );
}

void KrossWordPuzzle::loadFile( const KUrl &url, Crossword::KrossWord::FileFormat fileFormat,
				bool loadCrashedFile ) {
    m_mainLibrary->statusBar()->showMessage( i18n("Loading...") );
    m_loadProgressDialog = createLoadProgressDialog();
    m_loadProgressDialog->show();

    m_mainCrossword->loadFile( url, fileFormat, loadCrashedFile );
}

KrossWordPuzzle::~KrossWordPuzzle() {
// /*    if ( m_previewImageGeneratorThread ) {
// 	disconnect( m_previewImageGeneratorThread,
// 	     SIGNAL(finished()), this, SLOT(previewImageGeneratorThreadFinished()) );
//
// 	m_previewImageGeneratorThread->terminate();
// 	m_previewImageGeneratorThread->wait();
// 	delete m_previewImageGeneratorThread;
//     }*/
}

void KrossWordPuzzle::closeEvent( QCloseEvent* event ) {
    if ( m_mainCrossword->isModified() ) {
	QString msg = i18n("The current crossword has been modified.\n"
			"Would you like to save it?" );
	int result = KMessageBox::warningYesNoCancel( this, msg, i18n("Close Document"),
			    KStandardGuiItem::save(), KStandardGuiItem::discard() );
	if ( result == KMessageBox::Cancel || (result == KMessageBox::Yes
			    && !m_mainCrossword->save()) )
	    event->setAccepted( false );
	else
	    event->accept();
    } else
	event->accept();

    // Closing normally, so the temporary file can be deleted
    if ( event->isAccepted() ) {
		m_mainCrossword->removeTempFile();
		m_mainCrossword->setState( CrossWordXmlGuiWindow::ShowingNothing ); // Cleanup
    }
}

void KrossWordPuzzle::dragEnterEvent( QDragEnterEvent *event ) {
    if ( event->mimeData()->hasUrls() )
	event->accept();
}

void KrossWordPuzzle::dropEvent( QDropEvent *event ) {
    if ( event->mimeData()->hasUrls() ) {
		QPoint pt = ui_start_new.libraryTree->mapFrom( this, event->pos() );
		QModelIndex index = ui_start_new.libraryTree->indexAt( pt );
		if ( index.isValid() ) {
			if ( index.data(Qt::UserRole + 2).toBool() || index.parent().isValid() ) { // dropped onto library folder
				QString subFolder = index.parent().isValid()
					? QFileInfo( index.parent().data(Qt::UserRole).toString() ).fileName()
					: QFileInfo( index.data(Qt::UserRole).toString() ).fileName();
				m_mainLibrary->libraryAddCrossword( event->mimeData()->urls(), subFolder );
			} else
				m_mainLibrary->libraryAddCrossword( event->mimeData()->urls() );
		} else {
			QList<QUrl> urls = event->mimeData()->urls();
			loadFile( urls.first() );
		}
    }
}

KDialog* KrossWordPuzzle::createLoadProgressDialog() {
    KDialog *dialog = new KDialog( this );
    dialog->setButtons( KDialog::None );
    dialog->setAttribute( Qt::WA_DeleteOnClose );
    // TODO: No max/min buttons
//     dialog->se
    dialog->setWindowTitle( i18n("Loading...") );
//     QWidget *loadDlg = new QWidget;
    QLabel *lblLoad = new QLabel( i18n("Loading the crossword, please wait...") );
//     ui_load_progress.setupUi( loadDlg );

    dialog->setMainWidget( lblLoad );
//     dialog->setMainWidget( loadDlg );
    dialog->setModal( true );

    return dialog;
}
/*
void KrossWordPuzzle::recentFileListContextMenuRequested( const QPoint &pos ) {
    QMenu *menu = popupMenuRecentFilesList();

    action( actionName(RecentTab_RecentFilesRemove) )->setEnabled( ui_start_new.recentFiles->itemAt(pos) );
    menu->exec( ui_start_new.recentFiles->mapToGlobal(pos) );
}*/
/*
void KrossWordPuzzle::recentFilesClearSlot() {
    ui_start_new.recentFiles->clear();
    m_recentFilesAction->clear();
    m_recentFilesAction->saveEntries( Settings::self()->config()->group("") );
}*/
/*
void KrossWordPuzzle::recentFilesRemoveSlot() {
    QListWidgetItem *item = ui_start_new.recentFiles->currentItem();
    if ( !item ) {
	kDebug() << "No current item";
	return;
    }

    // Extract url from display string, because QListWidgetItem doesn't store Qt::UserRole-data..
    QString url = item->text().remove(QRegExp("(^.*\\(|\\)$)"));
    delete ui_start_new.recentFiles->takeItem( ui_start_new.recentFiles->row(item) );
    m_recentFilesAction->removeUrl( KUrl(url) );
    m_recentFilesAction->saveEntries( Settings::self()->config()->group("") );
}*/
/*
void KrossWordPuzzle::loadRecentItem() {
    QListWidgetItem *item = ui_start_new.recentFiles->currentItem();
    if ( !item )
	return;

    // Extract url from display string, because QListWidgetItem doesn't store Qt::UserRole-data..
    QString url = item->text().remove(QRegExp("(^.*\\(|\\)$)"));
    loadSlot( KUrl(url) );
}

void KrossWordPuzzle::recentFileExecuted( QListWidgetItem *item ) {
    if ( !item )
	return;

    // Extract url from display string, because QListWidgetItem doesn't store Qt::UserRole-data..
    QString url = item->text().remove(QRegExp("(^.*\\(|\\)$)"));
    loadSlot( KUrl(url) );
}*/

void KrossWordPuzzle::setupActions() {
    KStandardAction::showStatusbar( this, SLOT(showStatusbarGlobal(bool)),
				    actionCollection() );
    KStandardAction::keyBindings( this, SLOT(configureShortcutsGlobal()),
				  actionCollection() );
    KStandardAction::configureToolbars( this, SLOT(configureToolbarsGlobal()),
					actionCollection() );

    KStandardAction::preferences( this, SLOT(optionsPreferencesSlot()),
				  actionCollection() );
//     KStandardGameAction::load(this, SLOT(loadSlot()), actionCollection())->setStatusTip( i18n("Load a crossword from a file") );

//     KAction *downloadAction = new KAction( KIcon("get-hot-new-stuff"),
// 					   i18n("Get new crosswords..."), actionCollection() );
//     downloadAction->setStatusTip( i18n("Download crosswords from other users.") );
//     actionCollection()->addAction( actionName(Game_Download), downloadAction );
//     connect( downloadAction, SIGNAL(triggered()), this, SLOT(downloadSlot()) );
//
//     KAction *uploadAction = new KAction( KIcon("network-server"),
// 					   i18n("Upload current crossword..."), actionCollection() );
//     uploadAction->setStatusTip( i18n("Share the current crossword with other users.") );
//     actionCollection()->addAction( actionName(Game_Upload), uploadAction );
//     connect( uploadAction, SIGNAL(triggered()), this, SLOT(uploadSlot()) );
/*
    m_recentFilesAction = KStandardGameAction::loadRecent(
	    this, SLOT(loadRecentSlot(KUrl)), actionCollection());
    m_recentFilesAction->setIcon( KIcon("document-open-recent") ); // Not set by KStandardAction...
    m_recentFilesAction->loadEntries( Settings::self()->config()->group("") );
    m_recentFilesAction->setStatusTip( i18n("Load recent crosswords") );*/
    //     KStandardAction::openNew(this, SLOT(fileNew()), actionCollection());
//     KStandardGameAction::quit(qApp, SLOT(closeAllWindows()), actionCollection())->setStatusTip( i18n("Quit the game") );
    //     KStandardAction::preferences(this, SLOT(optionsPreferences()), actionCollection());
/*
    m_undoStack->createUndoAction( actionCollection(), actionName(Edit_Undo) );
    m_undoStack->createRedoAction( actionCollection(), actionName(Edit_Redo) );*/
/*
    KStandardGameAction::gameNew(this, SLOT(gameNewSlot()), actionCollection())->setStatusTip( i18n("Start a new crossword") );*/
}

void KrossWordPuzzle::gameNewSlot() {
    // create a new window
    (new KrossWordPuzzle)->show();

//     setState( ShowingStartPage );
//     setCurrentFileName( QString() );
}

void KrossWordPuzzle::downloadSlot() {
    KNS::Entry::List entries = KNS::Engine::download();
    kDebug() << "Entries count =" << entries.count();
//     KNS::Engine engine( this );
//     if ( engine.init("krosswordpuzzle.knsrc" )) {
// 	KNS::Entry::List entries = engine.downloadDialogModal( this );
//
// 	kDebug() << "Entries count =" << entries.count();
// 	if (entries.size() > 0) {
// 	    foreach ( KNS::Entry *entry, entries ) {
// 		// Downloaded file has the name "hotstuff-access" which is wrong (maybe it works
// 		// better with archives). So rename the file to the right name from the payload:
// 		QString filename = entry->payload().representation()
// 		    .remove( QRegExp("^.*\\?file=") ).remove( QRegExp("&site=.*$") );
// 		QStringList installedFiles = entry->installedFiles();
//
// 		kDebug() << "installedFiles =" << installedFiles;
// // 		if ( !installedFiles.isEmpty() ) {
// // 		    QString installedFile = installedFiles[0];
// //
// // 		    QString path = KUrl( installedFile ).path().remove( QRegExp("/[^/]*$") ) + "/";
// // 		    QFile( installedFile ).rename( path + filename );
//
// // 		    qDebug() << "PublicTransportSettings::downloadServiceProvidersClicked" <<
// // 		    "Rename" << installedFile << "to" << path + filename;
// // 		}
// 	    }
// 	}
//     }
}

void KrossWordPuzzle::uploadSlot() {
    // TODO: Get filename from CrossWordXmlGuiWindow
//     KNS::Entry *entry = KNS::Engine::upload( m_curFileName );
//     if ( entry )
// 	kDebug() << "Entry =" << entry->payload().translated( entry->payload().language() );
//     else
// 	kDebug() << "No uploaded entry";
}

void KrossWordPuzzle::loadSlot( const KUrl &url ) {
    loadFile( url );
}

void KrossWordPuzzle::loadRecentSlot( const KUrl& url ) {
    loadFile( url );
}

void KrossWordPuzzle::saveSlot() {
    m_mainCrossword->save();
}

void KrossWordPuzzle::saveAsSlot() {
    m_mainCrossword->saveAs();
}

void KrossWordPuzzle::optionsPreferencesSlot() {
  // Compare the names of the widgets in the .ui file
  // to the names of the variables in the .kcfg file.
  
  // Avoid to have 2 dialogs shown
  if ( KConfigDialog::showDialog("settings") )  {
    return;
  }
  
  KConfigDialog *dialog = new KConfigDialog( this, "settings", Settings::self() );

#if QT_VERSION >= 0x040600
  QWidget *animationSettingsDlg = new QWidget;
  ui_settings.setupUi( animationSettingsDlg );
  dialog->addPage( animationSettingsDlg, i18n("Animations"), "package_settings" );
#endif

  QWidget *themeSelectorDlg = new QWidget;
  KGameThemeSelector *themeSelector = new KGameThemeSelector( themeSelectorDlg,
		Settings::self(), KGameThemeSelector::NewStuffDisableDownload );
  dialog->addPage( themeSelector, i18n("Theme"), "games-config-theme" );
  
  connect( dialog, SIGNAL(settingsChanged(QString)),
	   this, SLOT(settingsChanged()) );
  dialog->setAttribute( Qt::WA_DeleteOnClose );
  dialog->show();
}

void KrossWordPuzzle::settingsChanged() {
  m_mainCrossword->updateTheme();
  
  m_mainCrossword->krossWord()->setAnimationTypes( animationTypesFromSettings() );
  if ( m_mainCrossword->solutionKrossWord() )
    m_mainCrossword->solutionKrossWord()->setAnimationTypes( animationTypesFromSettings() );
}

AnimationTypes KrossWordPuzzle::animationTypesFromSettings() {
  if ( !Settings::animate() )
    return NoAnimation;
  
  AnimationTypes anim = NoAnimation;
  if ( Settings::animateSizeChange() )
    anim |= AnimateSizeChange;
  if ( Settings::animatePosChange() )
    anim |= AnimatePosChange;
  if ( Settings::animateAppear() )
    anim |= AnimateAppear;
  if ( Settings::animateDisappear() )
    anim |= AnimateDisappear;
  if ( Settings::animateChangeLetter() )
    anim |= AnimateChangeLetter;
  if ( Settings::animateFocusIn() )
    anim |= AnimateFocusIn;
  if ( Settings::animateTransition() )
    anim |= AnimateTransition;

  return anim;
}

QMenu* KrossWordPuzzle::popupMenuRecentFilesList() {
    return static_cast<QMenu*>( factory()->container("recent_files_list_popup", this) );
}

#include "krosswordpuzzle.moc"